from abc import ABC, abstractmethod
from typing import List, Iterable, Optional, Set, Dict

from pydantic import BaseModel, Field

from mahjong_utils.models.furo import Furo, Kan
from mahjong_utils.models.mentsu import Mentsu, Shuntsu, Kotsu
from mahjong_utils.models.tatsu import Tatsu
from mahjong_utils.models.tile import Tile


class Hand(BaseModel, ABC):
    """
    表示一个结构分析后的手牌
    """

    @property
    @abstractmethod
    def tiles(self) -> Iterable[Tile]:
        raise NotImplementedError()

    @property
    @abstractmethod
    def menzen(self) -> bool:
        raise NotImplementedError()

    with_got: bool = False
    shanten: Optional[int] = None
    advance: Optional[Set[Tile]] = None
    discard_to_advance: Optional[Dict[Tile, Set[Tile]]] = None


class RegularHand(Hand):
    """
    表示一个结构分析后的以标准形为目标的手牌
    """

    k: int
    jyantou: Optional[Tile] = None
    menzen_mentsu: List[Mentsu] = Field(default_factory=list)
    furo: List[Furo] = Field(default_factory=list)
    tatsu: List[Tatsu] = Field(default_factory=list)
    remaining: List[Tile] = Field(default_factory=list)

    @property
    def mentsu(self) -> Iterable[Mentsu]:
        for mt in self.menzen_mentsu:
            yield mt
        for fr in self.furo:
            yield fr

    @property
    def shuntsu(self) -> Iterable[Shuntsu]:
        for mt in self.mentsu:
            if isinstance(mt, Shuntsu):
                yield mt

    @property
    def kotsu(self) -> Iterable[Kotsu]:
        for mt in self.mentsu:
            if isinstance(mt, Kotsu):
                yield mt

    @property
    def anko(self) -> Iterable[Kotsu]:
        for mt in self.menzen_mentsu:
            if isinstance(mt, Kotsu):
                yield mt
        for fr in self.furo:
            if isinstance(fr, Kan) and fr.ankan:
                yield fr

    @property
    def tiles(self) -> Iterable[Tile]:
        if self.jyantou is not None:
            yield self.jyantou
            yield self.jyantou
        for mt in self.mentsu:
            for t in mt.tiles:
                yield t
        for tt in self.tatsu:
            yield tt.first
            yield tt.second
        for t in self.remaining:
            yield t

    @property
    def menzen(self) -> bool:
        for fr in self.furo:
            if not (isinstance(fr, Kan) and fr.ankan):
                return False
        return True


class ChitoiHand(Hand):
    """
    表示一个结构分析后的以七对子为目标的手牌
    """

    pairs: List[Tile] = Field(default_factory=list)
    remaining: List[Tile] = Field(default_factory=list)

    @property
    def tiles(self) -> Iterable[Tile]:
        for p in self.pairs:
            yield p
            yield p
        for t in self.remaining:
            yield t

    @property
    def menzen(self) -> bool:
        return True


class KokushiHand(Hand):
    """
    表示一个结构分析后的以国士无双为目标的手牌
    """

    yaochu: List[Tile] = Field(default_factory=list)
    repeated: Optional[Tile] = None
    remaining: List[Tile] = Field(default_factory=list)

    @property
    def tiles(self) -> Iterable[Tile]:
        for t in self.yaochu:
            if self.repeated == t:
                yield t
            yield t
        for t in self.remaining:
            yield t

    @property
    def menzen(self) -> bool:
        return True


__all__ = ("Hand", "RegularHand", "ChitoiHand", "KokushiHand")



from typing import Callable, NamedTuple, Optional

from mahjong_utils.models.hora_hand import HoraHand


class Yaku(NamedTuple):
    name: str
    han: int
    furo_loss: int
    is_yakuman: bool = False
    checker: Optional[Callable[[HoraHand], bool]] = None

    def __call__(self, hora_hand: "HoraHand") -> bool:
        if self.checker:
            return self.checker(hora_hand)
        else:
            return False

    def __hash__(self):
        return hash(self.name)


def _yaku(han: int, furo_loss: int, is_yakuman: bool = False):
    def decorator(func):
        return Yaku(func.__name__, han, furo_loss, is_yakuman, func)

    return decorator


__all__ = ("Yaku", "_yaku")



def next_k_bit_number(x: int) -> int:
    # REF: http://realtimecollisiondetection.net/blog/?p=78
    b = x & -x
    t = x + b
    c = x ^ t
    m = (c >> 2) // b
    r = t | m
    return r


def generate_k_bit_number(k: int):
    x = (1 << k) - 1
    while True:
        yield x
        x = next_k_bit_number(x)



from typing import Iterable, Set

from mahjong_utils.internal.mentsu_utils import mentsu_try_exclude_got
from mahjong_utils.internal.tile_cling import tile_cling
from mahjong_utils.models.hand import RegularHand
from mahjong_utils.models.tile import Tile


def calc_regular_shanten(hand: RegularHand):
    """
    计算向听数

    :param hand:
    :return:
    """
    shanten = 2 * (hand.k - len(hand.menzen_mentsu)) - len(hand.tatsu)
    if hand.jyantou is not None:
        shanten -= 1
    return shanten


def calc_regular_advance(hand: RegularHand) -> Set[Tile]:
    """
    计算进张

    :param hand:
    :return:
    """
    advance = set()

    # 搭子的进张
    for tt in hand.tatsu:
        advance |= tt.waiting

    # 浮张的靠张
    if len(hand.furo) + len(hand.menzen_mentsu) + len(hand.tatsu) < hand.k:
        for t in hand.remaining:
            advance |= tile_cling[t]

    # 无雀头
    if hand.jyantou is None:
        for t in hand.remaining:
            advance.add(t)

    return advance


def hand_exclude_got_regular(hand: RegularHand, got: Tile) -> Iterable[RegularHand]:
    """
    hand为已摸牌状态的手牌，获取所有可能的未摸牌状态的手牌
    :param hand:
    :param got:
    :return:
    """
    assert hand.with_got

    # 扣掉雀头
    if hand.jyantou == got:
        new_hand = hand.copy(update={
            "with_got": False,
            "jyantou": None,
            "remaining": hand.remaining + [got]
        })
        yield new_hand

    # 扣掉面子
    for i, mt in enumerate(hand.menzen_mentsu):
        tatsu = mentsu_try_exclude_got(mt, got)
        if tatsu is not None:
            new_hand = hand.copy(update={
                "with_got": False,
                "menzen_mentsu": hand.menzen_mentsu[:i] + hand.menzen_mentsu[i + 1:],
                "tatsu": hand.tatsu + [tatsu]
            })
            yield new_hand

    # 扣掉搭子
    for i, tt in enumerate(hand.tatsu):
        if got == tt.first:
            new_hand = hand.copy(update={
                "with_got": False,
                "tatsu": hand.tatsu[:i] + hand.tatsu[i + 1:],
                "remaining": hand.remaining + [tt.second]
            })
            yield new_hand
        elif got == tt.second:
            new_hand = hand.copy(update={
                "with_got": False,
                "tatsu": hand.tatsu[:i] + hand.tatsu[i + 1:],
                "remaining": hand.remaining + [tt.first]
            })
            yield new_hand

    # 扣掉浮张
    try:
        idx = hand.remaining.index(got)
    except ValueError:
        idx = -1

    if idx != -1:
        new_hand = hand.copy(update={
            "with_got": False,
            "remaining": hand.remaining[:idx] + hand.remaining[idx + 1:]
        })
        yield new_hand



from enum import Enum


class TileType(str, Enum):
    """
    牌的花色
    """

    M = "m"
    P = "p"
    S = "s"
    Z = "z"

    @property
    def num_range(self):
        if self == TileType.Z:
            return range(1, 8)
        else:
            return range(0, 10)


__all__ = ("TileType",)



from abc import ABC
from typing import List
from typing import Optional, Iterable

from pydantic import Field, root_validator

from mahjong_utils.internal.hand_utils import hand_exclude_got_regular
from mahjong_utils.models.furo import Pon, Kan
from mahjong_utils.models.hand import RegularHand, Hand, ChitoiHand, KokushiHand
from mahjong_utils.models.mentsu import Kotsu
from mahjong_utils.models.tatsu import Penchan, Kanchan, Tatsu
from mahjong_utils.models.tile import Tile
from mahjong_utils.models.tile import all_yaochu
from mahjong_utils.models.tile import is_yaochu
from mahjong_utils.models.tile_type import TileType
from mahjong_utils.models.wind import Wind


class HoraHand(Hand, ABC):
    agari: Tile
    tsumo: bool
    hu: int
    self_wind: Optional[Wind]
    round_wind: Optional[Wind]


class RegularHoraHand(HoraHand, RegularHand):
    agari_tatsu: Optional[Tatsu]
    hu: int = 0  # 在__init__中计算

    k: int = 4
    jyantou: Tile

    @root_validator
    def validate_k(cls, values):
        assert len(values["menzen_mentsu"]) + len(values["furo"]) == 4
        return values

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.hu = self._calc_hu()

    def _calc_hu(self) -> int:
        ans = 20

        # 单骑、边张、坎张听牌
        if self.agari_tatsu is None or isinstance(self.agari_tatsu, Penchan) or isinstance(self.agari_tatsu, Kanchan):
            ans += 2

        # 明刻、杠
        for fr in self.furo:
            if isinstance(fr, Pon):
                if is_yaochu(fr.tile):
                    ans += 4
                else:
                    ans += 2
            elif isinstance(fr, Kan):
                if not fr.ankan:
                    if is_yaochu(fr.tile):
                        ans += 16
                    else:
                        ans += 8
                else:
                    if is_yaochu(fr.tile):
                        ans += 32
                    else:
                        ans += 16

        # 暗刻（不含暗杠）
        for mt in self.menzen_mentsu:
            if isinstance(mt, Kotsu):
                if is_yaochu(mt.tile):
                    ans += 8
                else:
                    ans += 4

        # 对碰荣和时该刻子计为明刻
        if isinstance(self.agari_tatsu, Kotsu):
            ans -= 2

        # 役牌雀头（连风算4符）
        if self.self_wind is not None and self.jyantou == self.self_wind.tile:
            ans += 2
        if self.round_wind is not None and self.jyantou == self.round_wind.tile:
            ans += 2
        if self.jyantou.tile_type == TileType.Z and self.jyantou.num >= 5:
            ans += 2

        # 门清荣和
        if self.menzen and not self.tsumo:
            ans += 10

        # 非平和自摸
        if ans != 20 and self.tsumo:
            ans += 2

        # 非门清最低30符
        if not self.menzen and ans < 30:
            ans = 30

        # 切上
        if ans % 10 != 0:
            ans += 10 - ans % 10

        return ans


class ChitoiHoraHand(HoraHand, ChitoiHand):
    hu: int = 25


class KokushiHoraHand(HoraHand, KokushiHand):
    repeated: Tile

    hu: int = 20

    yaochu: List[Tile] = Field(default_factory=lambda: list(all_yaochu))
    remaining: List[Tile] = Field(default_factory=list)

    @property
    def thirteen_waiting(self) -> bool:
        return self.agari == self.repeated


def _build_regular_hora_hand(hand: RegularHand,
                             agari: Tile,
                             tsumo: bool,
                             self_wind: Optional[Wind] = None,
                             round_wind: Optional[Wind] = None) -> Iterable[RegularHoraHand]:
    if hand.k != 4:
        raise ValueError("hand.k must be 4")

    if not hand.with_got:
        if hand.jyantou is not None:
            yield RegularHoraHand(agari=agari,
                                  tsumo=tsumo,
                                  self_wind=self_wind,
                                  round_wind=round_wind,
                                  agari_tatsu=hand.tatsu[0],
                                  jyantou=hand.jyantou,
                                  menzen_mentsu=hand.menzen_mentsu + [hand.tatsu[0].with_waiting(agari)],
                                  furo=hand.furo)
        else:
            yield RegularHoraHand(agari=agari,
                                  tsumo=tsumo,
                                  self_wind=self_wind,
                                  round_wind=round_wind,
                                  agari_tatsu=None,
                                  jyantou=agari,
                                  menzen_mentsu=hand.menzen_mentsu,
                                  furo=hand.furo)
    else:
        for excluded_hand in hand_exclude_got_regular(hand, agari):
            for h in _build_regular_hora_hand(excluded_hand, agari, tsumo, self_wind, round_wind):
                yield h


def _build_chitoi_hora_hand(hand: ChitoiHand,
                            agari: Tile,
                            tsumo: bool,
                            self_wind: Optional[Wind] = None,
                            round_wind: Optional[Wind] = None) -> Iterable[ChitoiHoraHand]:
    if not hand.with_got:
        if agari not in hand.advance:
            raise ValueError("agari is not waiting")

        pairs = hand.pairs.copy()
        pairs.append(agari)

        yield ChitoiHoraHand(agari=agari,
                             tsumo=tsumo,
                             self_wind=self_wind,
                             round_wind=round_wind,
                             pairs=pairs)
    else:
        if agari not in hand.pairs:
            raise ValueError("agari is not in hand")

        yield ChitoiHoraHand(agari=agari,
                             tsumo=tsumo,
                             self_wind=self_wind,
                             round_wind=round_wind,
                             pairs=hand.pairs)


def _build_kokushi_hora_hand(hand: KokushiHand,
                             agari: Tile,
                             tsumo: bool,
                             self_wind: Optional[Wind] = None,
                             round_wind: Optional[Wind] = None) -> Iterable[KokushiHoraHand]:
    if not hand.with_got:
        if agari not in hand.advance:
            raise ValueError("agari is not waiting")

        if hand.repeated is None:
            repeated = agari
        else:
            repeated = hand.repeated
    else:
        repeated = hand.repeated

    yield KokushiHoraHand(agari=agari,
                          tsumo=tsumo,
                          self_wind=self_wind,
                          round_wind=round_wind,
                          repeated=repeated)


def build_hora_hand(hand: Hand,
                    agari: Tile,
                    tsumo: bool,
                    self_wind: Optional[Wind] = None,
                    round_wind: Optional[Wind] = None) -> Iterable[HoraHand]:
    """
    根据给定Hand构造HoraHand。当Hand为摸牌状态时，返回所有可能的HoraHand。否则返回一个对应的HoraHand。

    :param hand: Hand
    :param agari: 和牌
    :param tsumo: 是否自摸
    :param self_wind: 自风
    :param round_wind: 场风
    :return: 当Hand为摸牌状态时，返回所有可能的HoraHand。否则返回一个对应的HoraHand。
    """
    if not hand.with_got:
        if hand.shanten != 0:
            raise ValueError("hand is not tenpai")
        if agari not in hand.advance:
            raise ValueError("agari is not waiting")
    else:
        if hand.shanten != -1:
            raise ValueError("hand is not agari")
        if agari not in hand.tiles:
            raise ValueError("agari is not in hand")

    if isinstance(hand, RegularHand):
        return _build_regular_hora_hand(hand, agari, tsumo, self_wind, round_wind)
    elif isinstance(hand, ChitoiHand):
        return _build_chitoi_hora_hand(hand, agari, tsumo, self_wind, round_wind)
    elif isinstance(hand, KokushiHand):
        return _build_kokushi_hora_hand(hand, agari, tsumo, self_wind, round_wind)
    else:
        raise TypeError(f"unexpected type of hand: {type(hand)}")



from typing import Optional, Union

from mahjong_utils.models.mentsu import Mentsu, Kotsu, Shuntsu
from mahjong_utils.models.tatsu import Tatsu, Toitsu, Penchan, Ryanmen, Kanchan
from mahjong_utils.models.tile import Tile


def kotsu_try_exclude_got(kotsu: Kotsu, got: Tile) -> Optional[Toitsu]:
    if got == kotsu.tile:
        return Toitsu(got)
    else:
        return None


def shuntsu_try_exclude_got(shuntsu: Shuntsu, got: Tile) -> Optional[Union[Penchan, Ryanmen, Kanchan]]:
    if got == shuntsu.tile:
        if got.num == 7:
            return Penchan(got + 1)
        else:
            return Ryanmen(got + 1)
    elif got == shuntsu.tile + 1:
        return Kanchan(shuntsu.tile)
    elif got == shuntsu.tile + 2:
        if got.num == 1:
            return Penchan(shuntsu.tile)
        else:
            return Ryanmen(shuntsu.tile)


def mentsu_try_exclude_got(mentsu: Mentsu, got: Tile) -> Optional[Tatsu]:
    if isinstance(mentsu, Kotsu):
        return kotsu_try_exclude_got(mentsu, got)
    elif isinstance(mentsu, Shuntsu):
        return shuntsu_try_exclude_got(mentsu, got)



from dataclasses import dataclass
from typing import List, Set, Dict, Literal, Optional, Mapping, Sequence

from mahjong_utils.internal.hand_utils import calc_regular_advance
from mahjong_utils.internal.legal_tiles_checker import ensure_legal_tiles
from mahjong_utils.internal.regular_hand_searcher import regular_hand_search
from mahjong_utils.models.furo import Furo
from mahjong_utils.models.hand import Hand, ChitoiHand, KokushiHand
from mahjong_utils.models.tile import Tile, is_yaochu, all_yaochu, tile


@dataclass
class ShantenResult:
    type: Literal['regular', 'chitoi', 'kokushi', 'union']
    shanten: int
    hands: Sequence[Hand]
    advance: Optional[Set[Tile]]
    discard_to_advance: Optional[Mapping[Tile, Set[Tile]]]


@dataclass
class UnionShantenResult(ShantenResult):
    regular: ShantenResult
    chitoi: Optional[ShantenResult]
    kokushi: Optional[ShantenResult]


# ======== 标准形 ========
def regular_shanten(tiles: List[Tile], furo: Optional[List[Furo]] = None) -> ShantenResult:
    ensure_legal_tiles(tiles)

    if furo is None:
        furo = []

    with_got = len(tiles) % 3 == 2
    k = len(tiles) // 3

    hands = regular_hand_search(k, tiles)

    for hand in hands:
        hand.with_got = with_got
        hand.furo = furo
        hand.k += len(furo)

    if with_got:
        advance_aggregated = None
        discard_to_advance_aggregated = dict()

        for hand in hands:
            hand.discard_to_advance = {}
            for i, discard in enumerate(hand.remaining):
                hand_after_discard = hand.copy(update={
                    "remaining": hand.remaining[0:i] + hand.remaining[i + 1:]
                })

                advance = calc_regular_advance(hand_after_discard)
                hand.discard_to_advance[discard] = advance

                if discard not in discard_to_advance_aggregated:
                    discard_to_advance_aggregated[discard] = advance
                else:
                    discard_to_advance_aggregated[discard] |= advance
    else:
        advance_aggregated = set()
        discard_to_advance_aggregated = None

        for hand in hands:
            hand.advance = calc_regular_advance(hand)
            advance_aggregated |= hand.advance

    result = ShantenResult(type="regular",
                           shanten=hands[0].shanten,
                           hands=hands,
                           advance=advance_aggregated,
                           discard_to_advance=discard_to_advance_aggregated)
    return result


# ======== 七对子 ========
def chitoi_shanten(tiles: List[Tile]) -> ShantenResult:
    ensure_legal_tiles(tiles, False)

    cnt: Dict[Tile, int] = {}
    for t in tiles:
        if t.num == 0:
            t = tile(t.tile_type, 5)
        cnt[t] = cnt.get(t, 0) + 1

    pairs = []
    remaining = []
    advance = set()
    for t in cnt:
        if cnt[t] >= 2:
            pairs.append(t)
            for i in range(cnt[t] - 2):
                remaining.append(t)
        elif cnt[t] == 1:
            advance.add(t)
            remaining.append(t)

    shanten = 6 - len(pairs)

    if len(tiles) == 13:
        hand = ChitoiHand(with_got=False,
                          shanten=shanten,
                          advance=advance,
                          pairs=pairs,
                          remaining=remaining)

        return ShantenResult(type="chitoi",
                             shanten=shanten,
                             hands=[hand],
                             advance=advance,
                             discard_to_advance=None)
    else:
        discard_to_advance = {}
        for t in remaining:
            advance_after_discard = advance.copy()
            if t in advance_after_discard:
                advance_after_discard.remove(t)
            discard_to_advance[t] = advance_after_discard

        hand = ChitoiHand(with_got=True,
                          shanten=shanten,
                          discard_to_advance=discard_to_advance,
                          pairs=pairs,
                          remaining=remaining)

        return ShantenResult(type="chitoi",
                             shanten=shanten,
                             hands=[hand],
                             advance=None,
                             discard_to_advance=discard_to_advance)


# ======== 国士无双 ========
def kokushi_shanten(tiles: List[Tile]) -> ShantenResult:
    ensure_legal_tiles(tiles, False)

    yaochu = set()
    repeated = set()
    remaining = []

    for t in tiles:
        if is_yaochu(t):
            if t in yaochu:
                if t in repeated:
                    remaining.append(t)
                else:
                    repeated.add(t)
            else:
                yaochu.add(t)
        else:
            remaining.append(t)

    if len(tiles) == 13:
        hands = []

        if len(repeated) > 0:
            shanten = 12 - len(yaochu)
            advance = all_yaochu - yaochu

            for t in repeated:
                hand = KokushiHand(with_got=False,
                                   shanten=shanten,
                                   advance=advance,
                                   yaochu=list(yaochu),
                                   repeated=t,
                                   remaining=[*remaining, *repeated.difference([t])])
                hands.append(hand)
        else:
            shanten = 13 - len(yaochu)
            advance = all_yaochu.copy()

            hand = KokushiHand(with_got=False,
                               shanten=shanten,
                               advance=advance,
                               yaochu=list(yaochu),
                               repeated=None,
                               remaining=remaining)
            hands.append(hand)

        return ShantenResult(type="kokushi",
                             shanten=shanten,
                             hands=hands,
                             advance=advance,
                             discard_to_advance=None)
    else:
        hands = []

        if len(repeated) > 0:
            shanten = 12 - len(yaochu)
            advance = all_yaochu - yaochu

            for t in repeated:
                discard_to_advance = {}
                for discard in remaining:
                    discard_to_advance[discard] = advance
                for discard in repeated:
                    if discard == t:
                        continue
                    discard_to_advance[discard] = advance

                hand = KokushiHand(with_got=True,
                                   shanten=shanten,
                                   discard_to_advance=discard_to_advance,
                                   yaochu=list(yaochu),
                                   repeated=t,
                                   remaining=[*remaining, *repeated.difference([t])])
                hands.append(hand)

            discard_to_advance = {}

            for h in hands:
                for discard, advance in h.discard_to_advance.items():
                    if discard not in discard_to_advance:
                        discard_to_advance[discard] = set()
                    discard_to_advance[discard] |= advance
        else:
            shanten = 13 - len(yaochu)
            advance = all_yaochu.copy()

            discard_to_advance = {}
            for discard in remaining:
                discard_to_advance[discard] = advance

            hand = KokushiHand(with_got=True,
                               shanten=shanten,
                               discard_to_advance=discard_to_advance,
                               yaochu=list(yaochu),
                               repeated=None,
                               remaining=remaining)
            hands.append(hand)

        return ShantenResult(type="kokushi",
                             shanten=shanten,
                             hands=hands,
                             advance=None,
                             discard_to_advance=discard_to_advance)


# ======== union ========
def shanten(tiles: List[Tile], furo: Optional[List[Furo]] = None) -> UnionShantenResult:
    ensure_legal_tiles(tiles)

    if furo is None:
        furo = []

    with_got = len(tiles) % 3 == 2
    k = len(tiles) // 3

    if k != 4:
        regular = regular_shanten(tiles, furo)
        return UnionShantenResult(type="union", shanten=regular.shanten, hands=regular.hands,
                                  advance=regular.advance, discard_to_advance=regular.discard_to_advance,
                                  regular=regular, chitoi=None, kokushi=None)

    regular = regular_shanten(tiles, furo)
    chitoi = chitoi_shanten(tiles)
    kokushi = kokushi_shanten(tiles)

    shanten = min(regular.shanten, chitoi.shanten, kokushi.shanten)
    hands = list()

    if not with_got:
        advance_aggregated = set()
        discard_to_advance_aggregated = None

        if regular.shanten == shanten:
            advance_aggregated |= regular.advance
            hands += regular.hands
        if chitoi.shanten == shanten:
            advance_aggregated |= chitoi.advance
            hands += chitoi.hands
        if kokushi.shanten == shanten:
            advance_aggregated |= kokushi.advance
            hands += kokushi.hands
    else:
        advance_aggregated = None
        discard_to_advance_aggregated = dict()

        if regular.shanten == shanten:
            hands += regular.hands
            for discard, advance in regular.discard_to_advance.items():
                if discard not in discard_to_advance_aggregated:
                    discard_to_advance_aggregated[discard] = set()
                discard_to_advance_aggregated[discard] |= advance
        if chitoi.shanten == shanten:
            hands += chitoi.hands
            for discard, advance in chitoi.discard_to_advance.items():
                if discard not in discard_to_advance_aggregated:
                    discard_to_advance_aggregated[discard] = set()
                discard_to_advance_aggregated[discard] |= advance
        if kokushi.shanten == shanten:
            hands += kokushi.hands
            for discard, advance in kokushi.discard_to_advance.items():
                if discard not in discard_to_advance_aggregated:
                    discard_to_advance_aggregated[discard] = set()
                discard_to_advance_aggregated[discard] |= advance

    return UnionShantenResult(type="union", shanten=shanten, hands=hands,
                              advance=advance_aggregated, discard_to_advance=discard_to_advance_aggregated,
                              regular=regular, chitoi=None, kokushi=None)


__all__ = ("regular_shanten",
           "chitoi_shanten",
           "kokushi_shanten",
           "shanten",
           "ShantenResult",
           "UnionShantenResult",)



from abc import ABC, abstractmethod
from typing import Set, Sequence, Union

from pydantic.dataclasses import dataclass

from .mentsu import Mentsu, Shuntsu, Kotsu
from .tile import Tile, parse_tiles, tiles_text
from .tile_type import TileType


@dataclass(frozen=True)
class Tatsu(ABC):
    """
    搭子
    """
    first: Tile

    @property
    @abstractmethod
    def second(self) -> Tile:
        raise NotImplementedError()

    def __repr__(self):
        return tiles_text([self.first, self.second])

    @property
    @abstractmethod
    def waiting(self) -> Set[Tile]:
        raise NotImplementedError()

    def with_waiting(self, tile: Tile) -> Mentsu:
        raise NotImplementedError()


@dataclass(frozen=True)
class Ryanmen(Tatsu):
    """
    两面
    """

    @property
    def second(self) -> Tile:
        return self.first + 1

    @property
    def waiting(self) -> Set[Tile]:
        return {self.first - 1, self.second + 1}

    def with_waiting(self, tile: Tile) -> Mentsu:
        if tile == self.first - 1:
            return Shuntsu(tile)
        elif tile == self.second + 1:
            return Shuntsu(self.first)
        else:
            raise ValueError(f"tile {tile} is not waiting")


@dataclass(frozen=True)
class Penchan(Tatsu):
    """
    边张
    """

    @property
    def second(self) -> Tile:
        return self.first + 1

    @property
    def waiting(self) -> Set[Tile]:
        if self.first.num == 1:
            return {self.first + 2}
        else:
            return {self.first - 1}

    def with_waiting(self, tile: Tile) -> Mentsu:
        if (self.first.num == 1 and tile == self.first + 2) or (self.first.num == 7 and tile == self.first - 1):
            return Shuntsu(self.first)
        else:
            raise ValueError(f"tile {tile} is not waiting")


@dataclass(frozen=True)
class Kanchan(Tatsu):
    """
    坎张
    """

    @property
    def second(self) -> Tile:
        return self.first + 2

    @property
    def waiting(self) -> Set[Tile]:
        return {self.first + 1}

    def with_waiting(self, tile: Tile) -> Mentsu:
        if tile == self.first + 1:
            return Shuntsu(self.first)
        else:
            raise ValueError(f"tile {tile} is not waiting")


@dataclass(frozen=True)
class Toitsu(Tatsu):
    """
    对子
    """

    @property
    def second(self) -> Tile:
        return self.first

    @property
    def waiting(self) -> Set[Tile]:
        return {self.first}

    def with_waiting(self, tile: Tile) -> Mentsu:
        if tile == self.first:
            return Kotsu(tile)
        else:
            raise ValueError(f"tile {tile} is not waiting")


def parse_tatsu(t: Union[Sequence[Tile], str]) -> Tatsu:
    if isinstance(t, str):
        t = parse_tiles(t)

    if len(t) != 2:
        raise ValueError("tiles must has length of 2")

    first, second = t[0], t[1]

    if first > second:
        first, second = second, first

    if first == second:
        return Toitsu(first)
    else:
        if first.tile_type == TileType.Z or second.tile_type == TileType.Z:
            raise ValueError(f"invalid tiles: {t}")

        if second - first == 1:
            if first.num == 1 or first.num == 7:
                return Penchan(first)
            else:
                return Ryanmen(first)
        elif second - first == 2:
            return Kanchan(first)
        else:
            raise ValueError(f"invalid tiles: {first}, {second}")


__all__ = ("Tatsu", "Toitsu", "Kanchan", "Ryanmen", "Penchan", "parse_tatsu")



from typing import List

from mahjong_utils.models.tile import Tile


def ensure_legal_tiles(tiles: List[Tile], allow_any_k: bool = True):
    if not allow_any_k and len(tiles) // 3 != 4:
        raise ValueError(f"invalid length of hand: {len(tiles)}")
    if len(tiles) < 1 or len(tiles) > 14 or len(tiles) % 3 == 0:
        raise ValueError(f"invalid length of hand: {len(tiles)}")

    cnt = {}
    for t in tiles:
        cnt[t] = cnt.get(t, 0) + 1
        if cnt[t] > 4:
            raise ValueError(f"invalid num of tile {t}")
